@@ -24,10 +24,12 @@
 #endif
 #include "gtkimhtml.h"
 #include <gtk/gtk.h>
+#include <glib/gerror.h>
 #include <gdk/gdkkeysyms.h>
 #include <string.h>
 #include <ctype.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <math.h>
 #ifdef HAVE_LANGINFO_CODESET
 #include <langinfo.h>
@@ -47,6 +49,19 @@
 #  define _(x) (x)
 #endif
 
+#include <pango/pango-font.h>
+
+#define TOOLTIP_TIMEOUT 500
+
+static gboolean gtk_motion_event_notify(GtkWidget *imhtml, GdkEventMotion *event, gpointer user_data);
+static gboolean gtk_leave_event_notify(GtkWidget *imhtml, GdkEventCrossing *event, gpointer user_data);
+
+static gboolean gtk_size_allocate_cb(GtkIMHtml *widget, GtkAllocation *alloc, gpointer user_data);
+
+static gboolean gaim_im_image_clicked(GtkWidget *w, GdkEvent *event, gaim_im_image *image);
+
+static gint gtk_imhtml_tip (gpointer data);
+
 
 /* POINT_SIZE converts from AIM font sizes to point sizes.  It probably should be redone in such a
  * way that it base the sizes off the default font size rather than using arbitrary font sizes. */
@@ -140,23 +155,28 @@ enum {
 };
 static guint signals [LAST_SIGNAL] = { 0 };
 
-static gboolean
-gtk_smiley_tree_destroy_from_hash(gpointer key, gpointer value,
-		gpointer user_data)
-{
-	gtk_smiley_tree_destroy(value);
-	return TRUE;
-}
-
 static void
 gtk_imhtml_finalize (GObject *object)
 {
 	GtkIMHtml *imhtml = GTK_IMHTML(object);
-	g_hash_table_foreach_remove(imhtml->smiley_data, gtk_smiley_tree_destroy_from_hash, NULL);
+	GList *scalables;
+
 	g_hash_table_destroy(imhtml->smiley_data);
 	gtk_smiley_tree_destroy(imhtml->default_smilies);
 	gdk_cursor_unref(imhtml->hand_cursor);
 	gdk_cursor_unref(imhtml->arrow_cursor);
+	if(imhtml->tip_window){
+		gtk_widget_destroy(imhtml->tip_window);
+	}
+	if(imhtml->tip_timer)
+		gtk_timeout_remove(imhtml->tip_timer);
+
+	for(scalables = imhtml->scalables; scalables; scalables = scalables->next) {
+		GtkIMHtmlScalable *scale = GTK_IMHTML_SCALABLE(scalables->data);
+		scale->free(scale);
+	}
+
+	g_list_free(imhtml->scalables);
 	G_OBJECT_CLASS(parent_class)->finalize (object);
 }
 
@@ -210,31 +230,46 @@ static void gtk_imhtml_init (GtkIMHtml *imhtml)
 	imhtml->show_smileys = TRUE;
 	imhtml->show_comments = TRUE;
 
-	imhtml->smiley_data = g_hash_table_new (g_str_hash, g_str_equal);
+	imhtml->smiley_data = g_hash_table_new_full(g_str_hash, g_str_equal,
+			g_free, (GDestroyNotify)gtk_smiley_tree_destroy);
 	imhtml->default_smilies = gtk_smiley_tree_new();
+
+	g_signal_connect(G_OBJECT(imhtml), "size-allocate", G_CALLBACK(gtk_size_allocate_cb), NULL);
+	g_signal_connect(G_OBJECT(imhtml), "motion-notify-event", G_CALLBACK(gtk_motion_event_notify), NULL);
+	g_signal_connect(G_OBJECT(imhtml), "leave-notify-event", G_CALLBACK(gtk_leave_event_notify), NULL);
+	gtk_widget_add_events(GTK_WIDGET(imhtml), GDK_LEAVE_NOTIFY_MASK);
+
+	imhtml->tip = NULL;
+	imhtml->tip_timer = 0;
+	imhtml->tip_window = NULL;
+
+	imhtml->scalables = NULL;
 }
 
 GtkWidget *gtk_imhtml_new(void *a, void *b)
 {
-	return GTK_WIDGET(gtk_type_new(gtk_imhtml_get_type()));
+	return GTK_WIDGET(g_object_new(gtk_imhtml_get_type(), NULL));
 }
 
-GtkType gtk_imhtml_get_type()
+GType gtk_imhtml_get_type()
 {
-	static guint imhtml_type = 0;
+	static GType imhtml_type = 0;
 
 	if (!imhtml_type) {
-		GtkTypeInfo imhtml_info = {
-			"GtkIMHtml",
-			sizeof (GtkIMHtml),
-			sizeof (GtkIMHtmlClass),
-			(GtkClassInitFunc) gtk_imhtml_class_init,
-			(GtkObjectInitFunc) gtk_imhtml_init,
+		static const GTypeInfo imhtml_info = {
+			sizeof(GtkIMHtmlClass),
 			NULL,
-			NULL
+			NULL,
+			(GClassInitFunc) gtk_imhtml_class_init,
+			NULL,
+			NULL,
+			sizeof (GtkIMHtml),
+			0,
+			(GInstanceInitFunc) gtk_imhtml_init
 		};
-		
-		imhtml_type = gtk_type_unique (gtk_text_view_get_type (), &imhtml_info);
+
+		imhtml_type = g_type_register_static(gtk_text_view_get_type(),
+				"GtkIMHtml", &imhtml_info, 0);
 	}
 
 	return imhtml_type;
@@ -280,28 +315,13 @@ gboolean tag_event(GtkTextTag *tag, GObject *arg1, GdkEvent *event, GtkTextIter
 			g_signal_emit(arg1, signals[URL_CLICKED], 0, url);
 			return FALSE;
 		} else if(event_button->button == 3) {
-			GtkWidget *img, *item, *label, *menu;
+			GtkWidget *img, *item, *menu;
 			struct url_data *tempdata = g_new(struct url_data, 1);
 			tempdata->object = g_object_ref(arg1);
 			tempdata->url = g_strdup(url);
-		
-			menu = gtk_menu_new();
-			gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, 
-							event_button->button, event_button->time);
-			
-			/* URL of link, desensitized */
-			item = gtk_menu_item_new();
-			label = gtk_label_new(url);
-			gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
-			gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
-			gtk_container_add(GTK_CONTAINER(item), label);
-			gtk_widget_set_sensitive(item, FALSE);
-			gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
 
-			/* seperator */
-			item = gtk_menu_item_new();
-			gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
-						
+			menu = gtk_menu_new();
+
 			/* buttons and such */
 			img = gtk_image_new_from_stock(GTK_STOCK_COPY, GTK_ICON_SIZE_MENU);
 			item = gtk_image_menu_item_new_with_mnemonic(_("_Copy Link Location"));
@@ -316,14 +336,13 @@ gboolean tag_event(GtkTextTag *tag, GObject *arg1, GdkEvent *event, GtkTextIter
 			g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(url_open),
 					tempdata);
 			gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
+
 			gtk_widget_show_all(menu);
-			
+			gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL,
+							event_button->button, event_button->time);
+
 			return TRUE;
 		}
-	} else if (event->type == GDK_ENTER_NOTIFY) {
-		/* make a hand cursor and a tooltip timeout -- if GTK worked as it should */
-	} else if (event->type == GDK_LEAVE_NOTIFY) {
-		/* clear timeout and make an arrow cursor again --if GTK worked as it should */
 	}
 	if(event->type == GDK_BUTTON_PRESS && event_button->button == 3)
 		return TRUE; /* Clicking the right mouse button on a link shouldn't
@@ -332,6 +351,71 @@ gboolean tag_event(GtkTextTag *tag, GObject *arg1, GdkEvent *event, GtkTextIter
 		return FALSE; /* Let clicks go through if we didn't catch anything */
 }
 
+gboolean gtk_motion_event_notify(GtkWidget *imhtml, GdkEventMotion *event, gpointer data)
+{	
+	GtkTextIter iter;
+	GdkWindow *win = event->window;
+	int x, y;
+	char *tip = NULL;
+	GSList *tags = NULL, *templist = NULL;
+	gdk_window_get_pointer(GTK_WIDGET(imhtml)->window, NULL, NULL, NULL);
+	gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(imhtml), GTK_TEXT_WINDOW_WIDGET,
+										  event->x, event->y, &x, &y);
+	gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(imhtml), &iter, x, y);
+	tags = gtk_text_iter_get_tags(&iter);
+
+	templist = tags;
+	while (templist) {
+		GtkTextTag *tag = templist->data;
+		tip = g_object_get_data(G_OBJECT(tag), "link_url");
+		if (tip)
+			break;
+		templist = templist->next;
+	}
+		
+	if (GTK_IMHTML(imhtml)->tip) {
+		if ((tip == GTK_IMHTML(imhtml)->tip)) {
+			return FALSE;
+		}
+		/* We've left the cell.  Remove the timeout and create a new one below */
+		if (GTK_IMHTML(imhtml)->tip_window) {
+			gtk_widget_destroy(GTK_IMHTML(imhtml)->tip_window);
+			GTK_IMHTML(imhtml)->tip_window = NULL;
+		}
+		gdk_window_set_cursor(win, GTK_IMHTML(imhtml)->arrow_cursor);
+		if (GTK_IMHTML(imhtml)->tip_timer)
+			g_source_remove(GTK_IMHTML(imhtml)->tip_timer);
+		GTK_IMHTML(imhtml)->tip_timer = 0;
+	}
+	
+	if(tip){
+		gdk_window_set_cursor(win, GTK_IMHTML(imhtml)->hand_cursor);
+		GTK_IMHTML(imhtml)->tip_timer = g_timeout_add (TOOLTIP_TIMEOUT, 
+							       gtk_imhtml_tip, imhtml);
+	}
+	
+	GTK_IMHTML(imhtml)->tip = tip;
+	g_slist_free(tags);
+	return FALSE;
+}
+
+gboolean gtk_leave_event_notify(GtkWidget *imhtml, GdkEventCrossing *event, gpointer data)
+{
+	/* when leaving the widget, clear any current & pending tooltips and restore the cursor */
+	if (GTK_IMHTML(imhtml)->tip_window) {
+		gtk_widget_destroy(GTK_IMHTML(imhtml)->tip_window);
+		GTK_IMHTML(imhtml)->tip_window = NULL;
+	}
+	if (GTK_IMHTML(imhtml)->tip_timer) {
+		g_source_remove(GTK_IMHTML(imhtml)->tip_timer);
+		GTK_IMHTML(imhtml)->tip_timer = 0;
+	}
+	gdk_window_set_cursor(event->window, GTK_IMHTML(imhtml)->arrow_cursor);
+
+	/* propogate the event normally */
+	return FALSE;
+}
+
 /* this isn't used yet
 static void
 gtk_smiley_tree_remove (GtkSmileyTree     *tree,
@@ -406,7 +490,7 @@ gtk_imhtml_associate_smiley (GtkIMHtml       *imhtml,
 	else if ((tree = g_hash_table_lookup(imhtml->smiley_data, sml))) {
 	} else {
 		tree = gtk_smiley_tree_new();
-		g_hash_table_insert(imhtml->smiley_data, sml, tree);
+		g_hash_table_insert(imhtml->smiley_data, g_strdup(sml), tree);
 	}
 
 	gtk_smiley_tree_insert (tree, smiley);
@@ -476,14 +560,14 @@ gtk_smiley_tree_image (GtkIMHtml     *imhtml,
 
 	return t->image->icon;
 }
-#define VALID_TAG(x)	if (!g_strncasecmp (string, x ">", strlen (x ">"))) {	\
+#define VALID_TAG(x)	if (!g_ascii_strncasecmp (string, x ">", strlen (x ">"))) {	\
 				*tag = g_strndup (string, strlen (x));		\
 				*len = strlen (x) + 1;				\
 				return TRUE;					\
 			}							\
 			(*type)++
 
-#define VALID_OPT_TAG(x)	if (!g_strncasecmp (string, x " ", strlen (x " "))) {	\
+#define VALID_OPT_TAG(x)	if (!g_ascii_strncasecmp (string, x " ", strlen (x " "))) {	\
 					const gchar *c = string + strlen (x " ");	\
 					gchar e = '"';					\
 					gboolean quote = FALSE;				\
@@ -517,25 +601,25 @@ gtk_imhtml_is_amp_escape (const gchar *string,
 	g_return_val_if_fail (replace != NULL, FALSE);
 	g_return_val_if_fail (length != NULL, FALSE);
 
-	if (!g_strncasecmp (string, "&amp;", 5)) {
+	if (!g_ascii_strncasecmp (string, "&amp;", 5)) {
 		*replace = '&';
 		*length = 5;
-	} else if (!g_strncasecmp (string, "&lt;", 4)) {
+	} else if (!g_ascii_strncasecmp (string, "&lt;", 4)) {
 		*replace = '<';
 		*length = 4;
-	} else if (!g_strncasecmp (string, "&gt;", 4)) {
+	} else if (!g_ascii_strncasecmp (string, "&gt;", 4)) {
 		*replace = '>';
 		*length = 4;
-	} else if (!g_strncasecmp (string, "&nbsp;", 6)) {
+	} else if (!g_ascii_strncasecmp (string, "&nbsp;", 6)) {
 		*replace = ' ';
 		*length = 6;
-	} else if (!g_strncasecmp (string, "&copy;", 6)) {
+	} else if (!g_ascii_strncasecmp (string, "&copy;", 6)) {
 		*replace = ''; /* was: '©' */
 		*length = 6;
-	} else if (!g_strncasecmp (string, "&quot;", 6)) {
+	} else if (!g_ascii_strncasecmp (string, "&quot;", 6)) {
 		*replace = '\"';
 		*length = 6;
-	} else if (!g_strncasecmp (string, "&reg;", 5)) {
+	} else if (!g_ascii_strncasecmp (string, "&reg;", 5)) {
 		*replace = ''; /* was: '®' */
 		*length = 5;
 	} else if (*(string + 1) == '#') {
@@ -618,7 +702,7 @@ gtk_imhtml_is_tag (const gchar *string,
 	VALID_OPT_TAG ("P");
 	VALID_OPT_TAG ("H3");
 
-	if (!g_strncasecmp(string, "!--", strlen ("!--"))) {
+	if (!g_ascii_strncasecmp(string, "!--", strlen ("!--"))) {
 		gchar *e = strstr (string + strlen("!--"), "-->");
 		if (e) {
 			*len = e - string + strlen ("-->");
@@ -637,7 +721,7 @@ gtk_imhtml_get_html_opt (gchar       *tag,
 	gchar *t = tag;
 	gchar *e, *a;
 
-	while (g_strncasecmp (t, opt, strlen (opt))) {
+	while (g_ascii_strncasecmp (t, opt, strlen (opt))) {
 		gboolean quote = FALSE;
 		if (*t == '\0') break;
 		while (*t && !((*t == ' ') && !quote)) {
@@ -648,7 +732,7 @@ gtk_imhtml_get_html_opt (gchar       *tag,
 		while (*t && (*t == ' ')) t++;
 	}
 
-	if (!g_strncasecmp (t, opt, strlen (opt))) {
+	if (!g_ascii_strncasecmp (t, opt, strlen (opt))) {
 		t += strlen (opt);
 	} else {
 		return NULL;
@@ -671,17 +755,17 @@ gtk_imhtml_get_html_opt (gchar       *tag,
 
 
 #define NEW_TEXT_BIT 0
-#define NEW_HR_BIT 1
 #define NEW_COMMENT_BIT 2
+#define NEW_SCALABLE_BIT 1
 #define NEW_BIT(x)	ws [wpos] = '\0'; \
                         mark2 = gtk_text_buffer_create_mark(imhtml->text_buffer, NULL, &iter, TRUE); \
                         gtk_text_buffer_insert(imhtml->text_buffer, &iter, ws, -1); \
-                      	gtk_text_buffer_get_end_iter(imhtml->text_buffer, &iter); \
+						gtk_text_buffer_get_end_iter(imhtml->text_buffer, &iter); \
                         gtk_text_buffer_get_iter_at_mark(imhtml->text_buffer, &siter, mark2); \
                         gtk_text_buffer_delete_mark(imhtml->text_buffer, mark2); \
                         if (bold) \
                                  gtk_text_buffer_apply_tag_by_name(imhtml->text_buffer, "BOLD", &siter, &iter); \
-		        if (italics) \
+						if (italics) \
                                  gtk_text_buffer_apply_tag_by_name(imhtml->text_buffer, "ITALICS", &siter, &iter); \
                         if (underline) \
                                  gtk_text_buffer_apply_tag_by_name(imhtml->text_buffer, "UNDERLINE", &siter, &iter); \
@@ -718,22 +802,26 @@ gtk_imhtml_get_html_opt (gchar       *tag,
                         } \
                         if (url) { \
                                  texttag = gtk_text_buffer_create_tag(imhtml->text_buffer, NULL, "foreground", "blue", "underline", PANGO_UNDERLINE_SINGLE, NULL); \
-                                 g_signal_connect(G_OBJECT(texttag), "event", G_CALLBACK(tag_event), strdup(url)); \
+                                 g_signal_connect(G_OBJECT(texttag), "event", G_CALLBACK(tag_event), g_strdup(url)); \
+                                 gtk_text_buffer_apply_tag(imhtml->text_buffer, texttag, &siter, &iter); \
+                                 texttag = gtk_text_buffer_create_tag(imhtml->text_buffer, NULL, NULL); \
+                                 g_object_set_data(G_OBJECT(texttag), "link_url", g_strdup(url)); \
                                  gtk_text_buffer_apply_tag(imhtml->text_buffer, texttag, &siter, &iter); \
                         } \
                         wpos = 0; \
                         ws[0] = 0; \
                         gtk_text_buffer_get_end_iter(imhtml->text_buffer, &iter); \
-                        if (x == NEW_HR_BIT) { \
-                                 GtkTextChildAnchor *anchor = gtk_text_buffer_create_child_anchor(imhtml->text_buffer, &iter); \
-                                 GtkWidget *sep = gtk_hseparator_new(); \
-                                 GdkRectangle rect; \
-                                 gtk_text_view_get_visible_rect(GTK_TEXT_VIEW(imhtml), &rect); \
-                                 gtk_widget_set_size_request(GTK_WIDGET(sep), rect.width, 2); \
-                                 gtk_text_view_add_child_at_anchor(GTK_TEXT_VIEW(imhtml), sep, anchor); \
-                                 gtk_widget_show(sep); \
+                        if (x == NEW_SCALABLE_BIT) { \
+								GdkRectangle rect; \
+								gtk_text_view_get_visible_rect(GTK_TEXT_VIEW(imhtml), &rect); \
+								scalable->add_to(scalable, imhtml, &iter); \
+								scalable->scale(scalable, rect.width, rect.height); \
+								imhtml->scalables = g_list_append(imhtml->scalables, scalable); \
+								gtk_text_buffer_get_end_iter(imhtml->text_buffer, &iter); \
                         } \
 
+
+
 GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 				 const gchar      *text,
 				 gint              len,
@@ -767,6 +855,8 @@ GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 	GdkRectangle rect;
 	int y, height;
 
+	GtkIMHtmlScalable *scalable = NULL;
+
 	g_return_val_if_fail (imhtml != NULL, NULL);
 	g_return_val_if_fail (GTK_IS_IMHTML (imhtml), NULL);
 	g_return_val_if_fail (text != NULL, NULL);
@@ -890,7 +980,8 @@ GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 				case 26:        /* HR */
 				case 42:        /* HR (opt) */
 					ws[wpos++] = '\n';
-					NEW_BIT(NEW_HR_BIT);
+					scalable = gaim_hr_new();
+					NEW_BIT(NEW_SCALABLE_BIT);
 					ws[wpos++] = '\n';
 					break;
 				case 27:	/* /FONT */
@@ -929,6 +1020,13 @@ GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 				case 39:	/* /HEAD */
 					break; 
 				case 40:        /* BINARY */
+					NEW_BIT (NEW_TEXT_BIT);
+					while (pos < len && g_ascii_strncasecmp("</BINARY>", c, strlen("</BINARY>"))) {
+						c++;
+						pos++;
+					}
+					c = c - tlen; /* Because it will add this later */
+					break;
 				case 41:        /* /BINARY */
 					break;
 				case 43:	/* FONT (opt) */
@@ -963,6 +1061,10 @@ GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 							font->face = face;
 						else if (oldfont && oldfont->face)
 							font->face = g_strdup(oldfont->face);
+						if (font->face && (atoi(font->face) > 100)) {
+							g_free(font->face);
+							font->face = g_strdup("100");
+						}
 
 						if (sml)
 							font->sml = sml;
@@ -1007,6 +1109,55 @@ GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 						}
 					}
 					break;
+				case 46:	/* IMG (opt) */
+					{
+						gchar *src = gtk_imhtml_get_html_opt (tag, "SRC=");
+						gchar *id = gtk_imhtml_get_html_opt (tag, "ID=");
+						gchar *datasize = gtk_imhtml_get_html_opt (tag, "DATASIZE=");
+						gint im_len = datasize?atoi(datasize):0;
+
+						if (src && id && im_len && im_len <= len - pos) {
+							/* This is an embedded IM image, or is it? */
+							char *tmp = NULL;
+							const char *alltext;
+							guchar *imagedata = NULL;
+
+							GdkPixbufLoader *load;
+							GdkPixbuf *imagepb = NULL;
+							GError *error = NULL;
+
+							tmp = g_strdup_printf("<DATA ID=\"%s\" SIZE=\"%s\">", id, datasize);
+							alltext = strstr(c, tmp);
+							imagedata = g_memdup(alltext + strlen(tmp), im_len);
+
+							g_free(tmp);
+
+							load = gdk_pixbuf_loader_new();
+							if (!gdk_pixbuf_loader_write(load, imagedata, im_len, &error)){
+								fprintf(stderr, "IM Image corrupted or unreadable.: %s\n", error->message);
+							} else {
+								imagepb = gdk_pixbuf_loader_get_pixbuf(load);
+								if (imagepb) {
+									scalable = gaim_im_image_new(imagepb, g_strdup(src));
+									NEW_BIT(NEW_SCALABLE_BIT);
+									g_object_unref(imagepb);
+								}
+							}
+
+							gdk_pixbuf_loader_close(load, NULL);
+
+
+							g_free(imagedata);
+							g_free(id);
+							g_free(datasize);
+							g_free(src);
+
+							break;
+						}
+						g_free(id);
+						g_free(datasize);
+						g_free(src);
+					}
 				case 47:	/* P (opt) */
 				case 48:	/* H3 (opt) */
 					break;
@@ -1115,6 +1266,8 @@ GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 		}
 	}
 	g_free (ws);
+	if(bg)
+		g_free(bg);
 	if (!(options & GTK_IMHTML_NO_SCROLL))
 		gtk_text_view_scroll_to_mark (GTK_TEXT_VIEW (imhtml), mark,
 					      0, TRUE, 0.0, 1.0);
@@ -1122,11 +1275,12 @@ GString* gtk_imhtml_append_text (GtkIMHtml        *imhtml,
 	return str;
 }
 
-void gtk_imhtml_remove_smileys(GtkIMHtml *imhtml) 
-{ 
+void gtk_imhtml_remove_smileys(GtkIMHtml *imhtml)
+{
 	g_hash_table_destroy(imhtml->smiley_data);
 	gtk_smiley_tree_destroy(imhtml->default_smilies);
-	imhtml->smiley_data = g_hash_table_new (g_str_hash, g_str_equal);
+	imhtml->smiley_data = g_hash_table_new_full(g_str_hash, g_str_equal,
+			g_free, (GDestroyNotify)gtk_smiley_tree_destroy);
 	imhtml->default_smilies = gtk_smiley_tree_new();
 }
 void       gtk_imhtml_show_smileys     (GtkIMHtml        *imhtml,
@@ -1156,3 +1310,341 @@ void gtk_imhtml_page_up (GtkIMHtml *imhtml)
 
 }
 void       gtk_imhtml_page_down        (GtkIMHtml        *imhtml){}
+
+static gint
+gtk_imhtml_tip_paint (GtkIMHtml *imhtml)
+{
+	PangoLayout *layout;
+
+	g_return_val_if_fail(GTK_IS_IMHTML(imhtml), FALSE);
+
+	layout = gtk_widget_create_pango_layout(imhtml->tip_window, imhtml->tip);
+
+	gtk_paint_flat_box (imhtml->tip_window->style, imhtml->tip_window->window, 
+						GTK_STATE_NORMAL, GTK_SHADOW_OUT, NULL, imhtml->tip_window,
+						"tooltip", 0, 0, -1, -1);
+
+	gtk_paint_layout (imhtml->tip_window->style, imhtml->tip_window->window, GTK_STATE_NORMAL,
+					  FALSE, NULL, imhtml->tip_window, NULL, 4, 4, layout);
+
+	g_object_unref(layout);
+	return FALSE;
+}
+
+static gint
+gtk_imhtml_tip (gpointer data)
+{
+	GtkIMHtml *imhtml = data;
+	PangoFontMetrics *font;
+	PangoLayout *layout;
+
+	gint gap, x, y, h, w, scr_w, baseline_skip;
+
+	g_return_val_if_fail(GTK_IS_IMHTML(imhtml), FALSE);
+
+	if (!imhtml->tip || !GTK_WIDGET_DRAWABLE (GTK_WIDGET(imhtml))) {
+		imhtml->tip_timer = 0;
+		return FALSE;
+	}
+	
+	if (imhtml->tip_window){
+		gtk_widget_destroy (imhtml->tip_window);
+		imhtml->tip_window = NULL;
+	}
+
+	imhtml->tip_timer = 0;
+	imhtml->tip_window = gtk_window_new (GTK_WINDOW_POPUP);
+	gtk_widget_set_app_paintable (imhtml->tip_window, TRUE);
+	gtk_window_set_resizable (GTK_WINDOW (imhtml->tip_window), FALSE);
+	gtk_widget_set_name (imhtml->tip_window, "gtk-tooltips");
+	g_signal_connect_swapped (G_OBJECT (imhtml->tip_window), "expose_event",
+							  G_CALLBACK (gtk_imhtml_tip_paint), imhtml);
+
+	gtk_widget_ensure_style (imhtml->tip_window);
+	layout = gtk_widget_create_pango_layout(imhtml->tip_window, imhtml->tip);
+	font = pango_font_get_metrics(pango_context_load_font(pango_layout_get_context(layout),
+														  imhtml->tip_window->style->font_desc),
+														  NULL);
+
+
+	pango_layout_get_pixel_size(layout, &scr_w, NULL);
+	gap = PANGO_PIXELS((pango_font_metrics_get_ascent(font) + 
+					   pango_font_metrics_get_descent(font))/ 4);
+
+	if (gap < 2)
+		gap = 2;
+	baseline_skip = PANGO_PIXELS(pango_font_metrics_get_ascent(font) + 
+								pango_font_metrics_get_descent(font));
+	w = 8 + scr_w;
+	h = 8 + baseline_skip;
+
+	gdk_window_get_pointer (NULL, &x, &y, NULL);
+	if (GTK_WIDGET_NO_WINDOW (GTK_WIDGET(imhtml)))
+		y += GTK_WIDGET(imhtml)->allocation.y;
+
+	scr_w = gdk_screen_width();
+
+	x -= ((w >> 1) + 4);
+
+	if ((x + w) > scr_w)
+		x -= (x + w) - scr_w;
+	else if (x < 0)
+		x = 0;
+
+	y = y + PANGO_PIXELS(pango_font_metrics_get_ascent(font) + 
+						pango_font_metrics_get_descent(font));
+
+	gtk_widget_set_size_request (imhtml->tip_window, w, h);
+	gtk_widget_show (imhtml->tip_window);
+	gtk_window_move (GTK_WINDOW(imhtml->tip_window), x, y);
+
+	pango_font_metrics_unref(font);
+	g_object_unref(layout);
+
+	return FALSE;
+}
+
+static gboolean gtk_size_allocate_cb(GtkIMHtml *widget, GtkAllocation *alloc, gpointer user_data)
+{
+	GdkRectangle rect;
+
+	gtk_text_view_get_visible_rect(GTK_TEXT_VIEW(widget), &rect);
+	if(widget->old_rect.width != rect.width || widget->old_rect.height != rect.height){
+		GList *iter = GTK_IMHTML(widget)->scalables;
+
+		while(iter){
+			GtkIMHtmlScalable *scale = GTK_IMHTML_SCALABLE(iter->data);
+			scale->scale(scale, rect.width, rect.height);
+
+			iter = iter->next;
+		}
+	}
+
+	widget->old_rect = rect;
+	return FALSE;
+}
+
+/* GtkIMHtmlScalable, gaim_im_image, gaim_hr */
+GtkIMHtmlScalable *gaim_im_image_new(GdkPixbuf *img, gchar *filename)
+{
+	gaim_im_image *im_image = g_malloc(sizeof(gaim_im_image));
+	GtkImage *image = GTK_IMAGE(gtk_image_new_from_pixbuf(img));
+
+	GTK_IMHTML_SCALABLE(im_image)->scale = gaim_im_image_scale;
+	GTK_IMHTML_SCALABLE(im_image)->add_to = gaim_im_image_add_to;
+	GTK_IMHTML_SCALABLE(im_image)->free = gaim_im_image_free;
+
+	im_image->pixbuf = img;
+	im_image->image = image;
+	im_image->width = gdk_pixbuf_get_width(img);
+	im_image->height = gdk_pixbuf_get_height(img);
+	im_image->mark = NULL;
+	im_image->filename = filename;
+
+	g_object_ref(img);
+	return GTK_IMHTML_SCALABLE(im_image);
+}
+
+void gaim_im_image_scale(GtkIMHtmlScalable *scale, int width, int height)
+{
+	gaim_im_image *image = (gaim_im_image *)scale;
+
+	if(image->width > width || image->height > height){
+		GdkPixbuf *new_image = NULL;
+		float factor;
+		int new_width = image->width, new_height = image->height;
+
+		if(image->width > width){
+			factor = (float)(width)/image->width;
+			new_width = width;
+			new_height = image->height * factor;
+		}
+		if(new_height > height){
+			factor = (float)(height)/new_height;
+			new_height = height;
+			new_width = new_width * factor;
+		}
+
+		new_image = gdk_pixbuf_scale_simple(image->pixbuf, new_width, new_height, GDK_INTERP_BILINEAR);
+		gtk_image_set_from_pixbuf(image->image, new_image);
+		g_object_unref(G_OBJECT(new_image));
+	}
+}
+
+void gaim_im_image_free(GtkIMHtmlScalable *scale)
+{
+	gaim_im_image *image = (gaim_im_image *)scale;
+
+	g_object_unref(image->pixbuf);
+	g_free(image->filename);
+	g_free(scale);
+}
+
+void gaim_im_image_add_to(GtkIMHtmlScalable *scale, GtkIMHtml *imhtml, GtkTextIter *iter)
+{
+	gaim_im_image *image = (gaim_im_image *)scale;
+	GtkWidget *box = gtk_event_box_new();
+	GtkTextChildAnchor *anchor = gtk_text_buffer_create_child_anchor(imhtml->text_buffer, iter);
+
+	gtk_container_add(GTK_CONTAINER(box), GTK_WIDGET(image->image));
+
+	gtk_widget_show(GTK_WIDGET(image->image));
+	gtk_widget_show(box);
+
+	gtk_text_view_add_child_at_anchor(GTK_TEXT_VIEW(imhtml), box, anchor);
+	g_signal_connect(G_OBJECT(box), "event", G_CALLBACK(gaim_im_image_clicked), image);
+}
+
+GtkIMHtmlScalable *gaim_hr_new()
+{
+	gaim_hr *hr = g_malloc(sizeof(gaim_hr));
+
+	GTK_IMHTML_SCALABLE(hr)->scale = gaim_hr_scale;
+	GTK_IMHTML_SCALABLE(hr)->add_to = gaim_hr_add_to;
+	GTK_IMHTML_SCALABLE(hr)->free = gaim_hr_free;
+
+	hr->sep = gtk_hseparator_new();
+	gtk_widget_set_size_request(hr->sep, 5000, 2);
+	gtk_widget_show(hr->sep);
+
+	return GTK_IMHTML_SCALABLE(hr);
+}
+
+void gaim_hr_scale(GtkIMHtmlScalable *scale, int width, int height)
+{
+	gtk_widget_set_size_request(((gaim_hr *)scale)->sep, width, 2);
+}
+
+void gaim_hr_add_to(GtkIMHtmlScalable *scale, GtkIMHtml *imhtml, GtkTextIter *iter)
+{
+	gaim_hr *hr = (gaim_hr *)scale;
+	GtkTextChildAnchor *anchor = gtk_text_buffer_create_child_anchor(imhtml->text_buffer, iter);
+
+	gtk_text_view_add_child_at_anchor(GTK_TEXT_VIEW(imhtml), hr->sep, anchor);
+}
+
+void gaim_hr_free(GtkIMHtmlScalable *scale)
+{
+/*	gtk_widget_destroy(((gaim_hr *)scale)->sep); */
+	g_free(scale);
+}
+
+static void write_img_to_file(GtkWidget *w, GtkFileSelection *sel)
+{
+	const gchar *filename = gtk_file_selection_get_filename(sel);
+	gaim_im_image *image = g_object_get_data(G_OBJECT(sel), "gaim_im_image");
+	gchar *type = NULL;
+	GError *error = NULL;
+#if GTK_CHECK_VERSION(2,2,0)
+	GSList *formats = gdk_pixbuf_get_formats();
+
+	while(formats){
+		GdkPixbufFormat *format = formats->data;
+		gchar **extensions = gdk_pixbuf_format_get_extensions(format);
+		gpointer p = extensions;
+
+		while(gdk_pixbuf_format_is_writable(format) && extensions && extensions[0]){
+			gchar *fmt_ext = extensions[0];
+			const gchar* file_ext = filename + strlen(filename) - strlen(fmt_ext);
+
+			if(!strcmp(fmt_ext, file_ext)){
+				type = gdk_pixbuf_format_get_name(format);
+				break;
+			}
+
+			extensions++;
+		}
+
+		g_strfreev(p);
+
+		if(type)
+			break;
+
+		formats = formats->next;
+	}
+
+	g_slist_free(formats);
+#else
+	/* this is really ugly code, but I think it will work */
+	char *basename = g_path_get_basename(filename);
+	char *ext = strrchr(basename, '.');
+
+	if(ext) {
+		ext++;
+		if(!g_ascii_strcasecmp(ext, "jpeg") || !g_ascii_strcasecmp(ext, "jpg"))
+			type = g_strdup("jpeg");
+		else if(!g_ascii_strcasecmp(ext, "png"))
+			type = g_strdup("png");
+	}
+
+	g_free(basename);
+#endif
+
+	/* If I can't find a valid type, I will just tell the user about it and then assume
+	   it's a png */
+	if(!type){
+		gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
+						_("Gaim was unable to guess the image type base on the file extension supplied.  Defaulting to PNG."));
+		type = g_strdup("png");
+	}
+
+	gdk_pixbuf_save(image->pixbuf, filename, type, &error, NULL);
+
+	if(error){
+		gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
+				_("Error saving image: %s"), error->message);
+		g_error_free(error);
+	}
+
+	g_free(type);
+}
+
+static void gaim_im_image_save(GtkWidget *w, gaim_im_image *image)
+{
+	GtkWidget *sel = gtk_file_selection_new(_("Gaim - Save Image"));
+
+	gtk_file_selection_set_filename(GTK_FILE_SELECTION(sel), image->filename);
+	g_object_set_data(G_OBJECT(sel), "gaim_im_image", image);
+	g_signal_connect(G_OBJECT(GTK_FILE_SELECTION(sel)->ok_button), "clicked",
+					 G_CALLBACK(write_img_to_file), sel);
+
+	g_signal_connect_swapped(G_OBJECT(GTK_FILE_SELECTION(sel)->ok_button), "clicked",
+							 G_CALLBACK(gtk_widget_destroy), sel);
+	g_signal_connect_swapped(G_OBJECT(GTK_FILE_SELECTION(sel)->cancel_button), "clicked",
+							 G_CALLBACK(gtk_widget_destroy), sel); 
+
+	gtk_widget_show(sel);
+}
+
+static gboolean gaim_im_image_clicked(GtkWidget *w, GdkEvent *event, gaim_im_image *image)
+{
+	GdkEventButton *event_button = (GdkEventButton *) event;
+
+	if (event->type == GDK_BUTTON_RELEASE) {
+		if(event_button->button == 3) {
+			GtkWidget *img, *item, *menu;
+			gchar *text = g_strdup_printf(_("_Save Image..."));
+			menu = gtk_menu_new();
+
+			/* buttons and such */
+			img = gtk_image_new_from_stock(GTK_STOCK_SAVE, GTK_ICON_SIZE_MENU);
+			item = gtk_image_menu_item_new_with_mnemonic(text);
+			gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(item), img);
+			g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(gaim_im_image_save), image);
+			gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
+
+			gtk_widget_show_all(menu);
+			gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL,
+							event_button->button, event_button->time);
+
+			g_free(text);
+			return TRUE;
+		}
+	}
+	if(event->type == GDK_BUTTON_PRESS && event_button->button == 3)
+		return TRUE; /* Clicking the right mouse button on a link shouldn't
+						be caught by the regular GtkTextView menu */
+	else
+		return FALSE; /* Let clicks go through if we didn't catch anything */
+
+}
